///File : main.cpp
//------------------------------------------------------------------------------
/*
    This file is part of Debtest.

    Debtest is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Debtest is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Debtest.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>


#include <stdio.h>
#include <stdlib.h>

#include <pthread.h>
//------------------------------------------------------------------------------
#include "config.hpp"
#include "main.hpp"
#include "common.hpp"
//------------------------------------------------------------------------------
using namespace std;
//------------------------------------------------------------------------------
void getPkgFilesList(vector<string> &pkgFilesList, progressStr &progress)
{
    pkgFilesList.clear();

    const vector<string> pkgList = getPkgList();

    for(size_t i = 0; i < pkgList.size(); i++)
    {
        editProgress(float(i) / float(pkgList.size()) * 100.f, progress);

        string listFilename = "/var/lib/dpkg/info/" + pkgList[i] + ".list";

        ifstream file(listFilename.c_str(), ios::in);  // on ouvre en lecture

        if(file)  // si l'ouverture a fonctionné
        {
            string currentFile;

            getline(file, currentFile); // On passe le "/."

            while(getline(file, currentFile))
            {
                pkgFilesList.push_back(currentFile);
            }

            file.close();
        }
        else
        {
            cerr << "File \"" << listFilename << "\" doesn't exist !" << endl;
            continue;
        }
    }

    editProgress(100, progress);
}
//------------------------------------------------------------------------------
void getAllFilesList(std::vector<std::vector<std::string> > &allFilesList, const std::vector<std::string> &excludePathList)
{
    allFilesList.clear();
    allFilesList.resize(1);

    // find / -path '/media' -prune -o -print
    string command = "find / ";
    for(int i = 0; i < excludePathList.size(); i++)
    {
        command += "-path '" + excludePathList[i] + "' -prune -o ";
    }
    command += "-print";

    FILE *pipe = popen(command.c_str(), "r");

    if(pipe)
    {
        string tmp;
        size_t containerId = 0;

        getLine(tmp, pipe); // On saute le dossier "/"

        while(getLine(tmp, pipe))
        {
            if(allFilesList[containerId].size() >= FILES_PER_CONTAINER)
            {
                containerId = allFilesList.size();
                allFilesList.resize(allFilesList.size() + 1);
            }

            allFilesList[containerId].push_back(tmp);
        }

        pclose(pipe);
    }
}
//------------------------------------------------------------------------------
void *threadSearch(void *data)
{
    threadDataStr &threadData = static_cast<threadDataStr&>(*static_cast<threadDataStr*>(data));

    int currentId = threadGetNewContainer(threadData);

    while(currentId != -1)
    {
        vector<string> &currentAllFilesList = threadData.allFilesList[currentId];
        const vector<string> &pkgFilesList = threadData.pkgFilesList;

        for(int i = currentAllFilesList.size() - 1; i >= 0; i--)
        {
            for(int j = 0; j < pkgFilesList.size(); j++)
            {
                if(currentAllFilesList[i] == pkgFilesList[j])
                {
                    currentAllFilesList.erase(currentAllFilesList.begin() + i);
                    break;
                }
            }
        }

        currentId = threadGetNewContainer(threadData);
    }
}
//------------------------------------------------------------------------------
int threadGetNewContainer(threadDataStr &threadData)
{
    int newId = 0;

    pthread_mutex_lock(&threadData.mutex);
    if(threadData.lastContainerId != -1) // Si pas déjà fini par un autre thread...
    {
        threadData.lastContainerId++;
    }

    if(threadData.lastContainerId >= threadData.allFilesList.size()) // Si fini, on l'indique pour les thread suivants.
    {
        threadData.lastContainerId = -1;
    }

    newId = threadData.lastContainerId;

    editProgress((float(newId) / float(threadData.allFilesList.size())) * 100.f, *threadData.progress);

    pthread_mutex_unlock(&threadData.mutex);

    return newId;
}
//------------------------------------------------------------------------------
void cmpFilesList(threadDataStr &threadData)
{
    pthread_mutex_init(&threadData.mutex, NULL);
    threadData.lastContainerId = 0;

    const int threadNb = threadData.progress->nbThread;
    pthread_t threadTbl[threadNb];

    // Creation
    for(int i = 0; i < threadNb; i++)
    {
        pthread_create(&threadTbl[i], NULL, threadSearch, (void*)&threadData);
    }

    // Attente
    for(int i = 0; i < threadNb; i++)
    {
        pthread_join(threadTbl[i], NULL);
    }
}
//------------------------------------------------------------------------------
void getCoords(const std::vector<std::vector<std::string> > &allFilesList, const size_t &id, size_t &i, size_t &j)
{
    size_t current = 0;

    for(i = 0; i < allFilesList.size(); i++)
    {
        if(id < current + allFilesList[i].size())
        {
            j = id - current;
            return;
        }
        current += allFilesList[i].size();
    }

    cerr << "Error : getCoords() ; id too big !" << endl;
    exit(-11);
}

//------------------------------------------------------------------------------
void filterResults(std::vector<std::vector<std::string> > &allFilesList, progressStr &progress)
{
    size_t x, y;
    vector<string>::iterator current, next;

    // Création du totale
    size_t total = 0;
    for(int i = 0; i < allFilesList.size(); i++)
    {
        total += allFilesList[i].size();
    }
    size_t backTotal = total;

    for(size_t i = 0; i < total - 1; i++)
    {
        editProgress(float((backTotal - total) + i) / float(backTotal) * 100.f, progress);

        getCoords(allFilesList, i, x, y);
        current = allFilesList[x].begin() + y;

        getCoords(allFilesList, i + 1, x, y);
        next = allFilesList[x].begin() + y;

        while(next->size() >= current->size() && next->substr(0, current->size()) == *current) // Quant on trouve un dossier contenant le fichier suivant, on continue l'analyse...
        {
            allFilesList[x].erase(next);
            total--;

            editProgress(float((backTotal - total) + i) / float(backTotal) * 100.f, progress);

            if(i < total - 1)
            {
                getCoords(allFilesList, i + 1, x, y);
                next = allFilesList[x].begin() + y;
            }
            else
            {
                break;
            }
        }
    }
}
//------------------------------------------------------------------------------
int getCoreNb()
{
    int coreNb = DEFAULT_THREAD_NB;

    FILE *pipe = popen("cat /proc/cpuinfo | grep processor | wc -l", "r");

    if(pipe)
    {
        string tmp;

        if(getLine(tmp, pipe))
        {
            std::stringstream ss;
            ss << tmp;
            ss >> coreNb;
        }

        pclose(pipe);
    }

    if(coreNb > 0)
    {
        return coreNb;
    }
    else
    {
        return DEFAULT_THREAD_NB;
    }
}
//------------------------------------------------------------------------------
int main(int argc, char *argv[])
{
    progressStr progress;
    pthread_mutex_init(&progress.mutex, NULL);
    progress.perCent = 0;
    //progress.verboseMode = VERBOSE_MODE_QUIET;//VERBOSE_MODE_NORMAL;
    progress.verboseMode = VERBOSE_MODE_NORMAL;
    progress.isFilter = true;
    progress.nbThread = getCoreNb();

    threadDataStr threadData;
    threadData.progress = &progress;

    // Exclude path
    vector<string> excludedPath;
    excludedPath.push_back("/boot");
    excludedPath.push_back("/cdrom");
    excludedPath.push_back("/dev");
    excludedPath.push_back("/etc");
    excludedPath.push_back("/home");
    excludedPath.push_back("/lost+found");
    excludedPath.push_back("/media");
    excludedPath.push_back("/mnt");
    excludedPath.push_back("/proc");
    excludedPath.push_back("/root");
    excludedPath.push_back("/selinux");
    excludedPath.push_back("/sys");
    excludedPath.push_back("/tmp");
    excludedPath.push_back("/var");

    excludedPath.push_back("/initrd.img");
    excludedPath.push_back("/initrd.img.old");
    excludedPath.push_back("/vmlinuz");
    excludedPath.push_back("/vmlinuz.old");



    // Args...
    bool isMnanualExcludedPath = false;
    for(int i = 1; i < argc; i++)
    {
        string arg = argv[i];

        if(arg == "-q" || arg == "--quiet")
        {
            progress.verboseMode = VERBOSE_MODE_QUIET;
        }
        else if(arg == "-V" || arg == "--version")
        {
            cout << "debtest " << VERSION << endl;
            return 0;
        }
        else if(arg.substr(0, 7) == "--jobs=")
        {
            int nb = 0;
            string nbStr = arg.substr(7);
            std::stringstream ss;
            ss << nbStr;
            ss >> nb;

            if(nb > 0)
            {
                progress.nbThread = nb;
            }
            else
            {
                cerr << "Error : \"" << nbStr << "\" is not a valide job number !" << endl;
                return 1;
            }
        }
        else if(arg == "--no-filter")
        {
            progress.isFilter = false;
            return 0;
        }
        else if(arg == "--filter")
        {
            progress.isFilter = true;
        }
        else if(arg == "-e" || arg == "--exclude" )
        {
            if(!isMnanualExcludedPath) // On efface les valeurs par default...
            {
                isMnanualExcludedPath = true;
                excludedPath.clear();
            }

            for( i++; i < argc ; i++)
            {
                arg = argv[i];

                if(arg[0] != '-') // Pas un param
                {
                    if(arg[0] != '/')
                    {
                        cerr << "Error : \"" << arg << "\" should be an absolute path !" << endl;
                        return 1;
                    }
                    else
                    {
                        excludedPath.push_back(arg);
                    }
                }
                else // Un param... On le traite comme les autres.
                {
                    i--; // Pour retraiter l'argument testé.
                    break;
                }
            }
        }

        else if(arg == "-h" || arg == "--help")
        {
            cout << "Debfilter will create a list of every files installed on your computer and compare it to the real file list. You must not delete directly the files given by this program: most of them are created by the packages and should not be removed !" << endl;
            cout << "General:" << endl;
            cout << "  -V,  --version           Print current version." << endl;
            cout << "  -h,  --help              Show help message." << endl;
            cout << endl;
            cout << "Interaction:" << endl;
            cout << "  -q,  --quiet             Don't show progress." << endl;
            cout << "  --filter                 Do filter results (When a directory in not present in the packages files, its content is not printed)." << endl;
            cout << "  --no-filter              Don't filter results (see --filter option)." << endl;
            cout << "  --jobs=jobs              Use \"jobs\" thread." << endl;
            cout << "  -x PATH..., --exclude PATH..." << endl;
            cout << "                           Exclude path \"PATH...\" from real file list." << endl;


            return 0;
        }
        else
        {
            cerr << "Error : argument \"" << arg << "\" unknown." << endl;
            return 1;
        }
    }

    // Excluded packages && nb cores
    if(progress.verboseMode != VERBOSE_MODE_QUIET)
    {
        cout << "These paths will be excluded : " << endl;
        for(int i = 0; i < excludedPath.size(); i++)
        {
            cout << "  " << excludedPath[i] << endl;
        }

        cout << "Thread number : " << progress.nbThread << endl;
    }

    // Progress thread
    pthread_t progressThread;
    if(progress.verboseMode != VERBOSE_MODE_QUIET)
    {
        pthread_create(&progressThread, NULL, progressFn, (void*)&progress);
    }

    // Packages list
    editProgressMsg("Creating deb files list...", progress);
    getPkgFilesList(threadData.pkgFilesList, progress);


    editProgressMsg("Creating real files list...", progress);
    getAllFilesList(threadData.allFilesList, excludedPath);

    editProgressMsg("Comparing files...", progress);
    cmpFilesList(threadData);

    if(progress.isFilter)
    {
        editProgressMsg("Filtering files...", progress);
        filterResults(threadData.allFilesList, progress);
    }

    // Saut de ligne + arret
    if(progress.verboseMode != VERBOSE_MODE_QUIET)
    {
        cout << endl;
        if(progress.verboseMode != VERBOSE_MODE_QUIET)
        {
            pthread_cancel(progressThread);
        }
    }

    // Result
    int total = 0;
    for(int i = 0; i < threadData.allFilesList.size(); i++)
    {
        total += threadData.allFilesList[i].size();
        for(int j = 0; j < threadData.allFilesList[i].size(); j++)
        {
            cout << threadData.allFilesList[i][j] << endl;
        }
    }

    if(progress.verboseMode != VERBOSE_MODE_QUIET)
    {
        cout << "Files found : " << total << endl;
    }

    return 0;
}
